/*
								+---------------------------------+
								|                                 |
								|      ***   The Life   ***       |
								|                                 |
								|    Copyright  L. Polok  2002   |
								|                                 |
								|            Life.cpp             |
								|                                 |
								+---------------------------------+
*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "Life.h"

static int die_till;  // cislo, kdy jeste bunka umre na "osameni" - default = 1
static int live_to;   // cislo, kdy bunka zemre na "nedostatek potravy" - default = 4
static int min_birth;
static int max_birth; // cislo (rozsah) kdy se narodi nova bunka - default = 3


// nahraje konstanty pro pocitani life
void Init_Life()
{
	FILE *fr;

	die_till = 1;
	live_to = 4;
	min_birth = 3;
	max_birth = 3;
	// konstanty - pocty sousedu :

	/*
		- Pro bunku, ktera je ziva :
			* Kazda bunka s mene nebo jednim sousedem chcipne
			* Kazda bunka s vice nez tremi sousedy chcipne
			* Kazda bunka se dvema nebo tremi sousedy prezije

		- Pro bunku, ktera je chcapla :
			* Kazda bunka se tremi sousedy se "narodi"

		(vymyslel to John Conway v roce 1970)
    */

	if((fr = fopen("Rules.ini", "r")) == NULL)
		return;

	fscanf(fr, "%d %d %d %d", &die_till, &live_to, &min_birth, &max_birth);
	// nacte konstanty ze souboru (nebo pouzije default)

	fclose(fr);
}

// vrati hodnotu bunky [x, y, z] v matrici p_m
// (nebo kdyz je xyz mimo tak hodnotu ze bunka neexistuje)
static __inline Cell Cell_Value(Life_Matrix *p_m, int x, int y, int z)
{
	if(x < 0 || x >= x_Size ||
	   y < 0 || y >= y_Size ||
	   z < 0 || z >= z_Size)
		return CELL_NONE;
	// zkontroluje hranice

	return (*p_m)[x][y][z];
}

// spocita zive sousedy bunky [x, y, z] v matrici p_m
static int Count_Neighbours(Life_Matrix *p_m, int x, int y, int z)
{
	int i, j, k, n;

	n = (Cell_Value(p_m, x, y, z) == CELL_LIVE)? -1 : 0;
	// algorytmus napocita i bunku samotnou

	for(i = x - 1; i <= x + 1; i ++) {
		for(j = y - 1; j <= y + 1; j ++) {
			for(k = z - 1; k <= z + 1; k ++) {
				if(Cell_Value(p_m, i, j, k) == CELL_LIVE)
					n ++;
			}
		}
	}
	// spocita sousedy bunky (max. 26)
	// (devet v rovine nad ni, devet
	// v rovine pod ni a osm okolo ni)

	return n;
}

// pocita life z prednastavenych konstant
// v matici p_m. bunky se rodi a umiraji
// podle poctu jejich sousedu
void Proceed_Life(Life_Matrix *p_m)
{
	Life_Matrix tmp;
	int x, y, z;
	int n_cnt;
	int i;

	memcpy(&tmp, p_m, sizeof(Life_Matrix));
	// zkopiruje matici do zasobniku

	for(i = 0; i < Max_Index; i ++) {
		x = (i & x_Mask) >> x_Shift;
		y = (i & y_Mask) >> y_Shift;
		z = (i & z_Mask) >> z_Shift;
		// spocita x, y, z - je to rychlejsi
		// nez tri cykly v sobe ;-)

		n_cnt = Count_Neighbours(&tmp, x, y, z);
		// spocita sousedy bunky

		if(tmp[x][y][z] == CELL_LIVE) {
			/*
				- Pro bunku, ktera je ziva :
					* Kazda bunka s mene nebo jednim sousedem chcipne
					* Kazda bunka s vice nez tremi sousedy chcipne
					* Kazda bunka se dvema nebo tremi sousedy prezije
			*/
			if(n_cnt <= die_till)
				(*p_m)[x][y][z] = CELL_DEAD;
			else if(n_cnt >= live_to)
				(*p_m)[x][y][z] = CELL_DEAD;
			else
				(*p_m)[x][y][z] = CELL_LIVE;
		} else {
			/*
				- Pro bunku, ktera je chcapla :
					* Kazda bunka se tremi sousedy se "narodi"
			*/
			if(n_cnt >= min_birth && n_cnt <= max_birth)
				(*p_m)[x][y][z] = CELL_LIVE;
			else
				(*p_m)[x][y][z] = CELL_DEAD;
		}
		// spocita co se s bunkou stane
	}
}

// vytvori matrici a ze souboru ini_file
// nahraje cisla reprezentujici pocet
// a pozice zivych bunek
Life_Matrix *Create_LMatrix(char *ini_file)
{
	Life_Matrix *tmp;
	int x, y, z;
	int l_num;
	FILE *fr;

	if((tmp = (Life_Matrix*) malloc(sizeof(Life_Matrix))) == NULL)
		return NULL;
	// naalokuje pamet pro life

	if((fr = fopen(ini_file, "r")) == NULL)
		return NULL;
	// otevre soubor

	memset(tmp, CELL_DEAD, sizeof(Life_Matrix));
	// nastavi vsechny bunky na chcaple

	fscanf(fr, "%d", &l_num);
	// nacte pocet zivych bunek

	for(; l_num > 0; l_num --) {
		fscanf(fr, " %d %d %d ", &x, &y, &z);
		// precte souradnice

		if(x < 0 || x >= x_Size ||
		   y < 0 || y >= y_Size ||
		   z < 0 || z >= z_Size)
			continue;
		// zkontroluje jestli se vejde..

		(*tmp)[x][y][z] = CELL_LIVE;
		// a nastavi bunku na zivou
	}

	fclose(fr);

	return tmp;
}



