/*
								+---------------------------------+
								|                                 |
								|        ***   Main   ***         |
								|                                 |
								|    Copyright  L. Polok  2002   |
								|                                 |
								|            Main.cpp             |
								|                                 |
								+---------------------------------+
*/

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <io.h>
#include "Directx.h"
#include "Matrix.h"
#include "Font_eng.h"
#include "Spline.h"
#include "Load_3ds.h"
#include "Raster.h"
#include "Life.h"
#include "Main.h"

#define APPNAME "S_Buffer_Engine"
#define WNDNAME "S_Buffer - Engine 3.0 (The game of life)"
static int Width = 640;
static int Height = 480;

unsigned __int32 *video;
unsigned __int32 *buffer;

#define PI 3.1415926

static World *cell;
// bunka

static Life_Matrix *svet;
// matice 64x64x64 ve ktere se simuluje life

static int auto_proceed = 0;
// pokud je 1, life se hraje sam

static Matrix cmatrix;
static Matrix _no_y_cmatrix;
static float y_angle;
static int mx, my, ax, ay;
// kamera + kramy pro mouselook

static int nFrame = 0;
// frame counter (pocitadlo snimku)

static int have_to_graph;
// urcuje ze se ma vytvorit graph.txt - spousti se "Home"

static int have_to_screenshot;
// pokud je 1, ma se udelat screenshot

static int bilerp = 1;
// urcuje zda bilinearni interpolace ma byt zapnuta / vypnuta

// vytvori soubor engine.log
// a vlozi do nej uvodni radku
void Eng_Log_Init()
{
	FILE *fw;
	fw = fopen("Engine.log", "w");
	fprintf(fw, "__________S_Buffer - Log__________\n");
	fclose(fw);
}

// pripoji k souboru engine.log text l_str a odradkuje
void Eng_Log(char *l_str)
{
	FILE *fw;
	fw = fopen("Engine.log", "a");
	fprintf(fw, "%s\n", l_str);
	fclose(fw);
}

void Make_ScreenShot(unsigned __int32 *scr)
{
	char str[256];
	BMP screen;
	int i;

	i = 0;
	do
		sprintf(str, "Screenshot%2x.bmp", i ++);
	while(_access(str, 0) != -1);
	// hleda index screenshotu ktery neexistuje
	// _access() je v <io.h>

	screen.n_Width = Width;
	screen.n_Height = Height;
	screen.bytes = scr;
	// vytvori bitmapu

	Save_TrueColor_BMP(str, &screen);
	// .. a ulozi ji
}

// main
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdline, int nCmdShow)
{
	static char szAppName[] = APPNAME;
	HWND		hwnd;
	MSG			msg;
	WNDCLASSEX	wndclass;
	int i;

	Eng_Log_Init();

	wndclass.cbSize			= sizeof (wndclass);
	wndclass.style			= CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc	= Wokno;
	wndclass.cbClsExtra		= 0;
	wndclass.cbWndExtra		= 0;
	wndclass.hInstance		= hInstance;
	wndclass.hIcon			= LoadIcon(hInstance, IDI_APPLICATION);
	wndclass.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground	= (HBRUSH) GetStockObject(BLACK_BRUSH);
	wndclass.lpszMenuName	= 0;
	wndclass.lpszClassName	= szAppName;
	wndclass.hIconSm		= LoadIcon(NULL, IDI_APPLICATION);

	RegisterClassEx(&wndclass);

	hwnd = CreateWindow(szAppName,
		WNDNAME, WS_DLGFRAME,
		CW_USEDEFAULT, CW_USEDEFAULT,
		Width, Height, NULL,
		NULL, hInstance, NULL);

	ShowWindow(hwnd, nCmdShow);
	UpdateWindow(hwnd);
	ShowCursor(FALSE);

	Matrix_Init(&_no_y_cmatrix);
	Matrix_Rotate_Z(&_no_y_cmatrix, (float)-PI);
	Matrix_Translate(&_no_y_cmatrix, 0, -30, -50);
	y_angle = 0;
	// kamera ok

	Eng_Log("Loading world ..");

	if((cell = Load_3DS("Cells\\Cell.3ds")) == NULL)
		return 0;
	for(i = 0; i < cell->n_object_num; i ++)
		Matrix_Init(&cell->object[i].m_object);
	// nahraje model "bunky"

	Eng_Log("    OK");

	Eng_Log("Loading textures ...");

	if(!Load_Textures(cell, "Cells"))
		return 0;
	// nahraje textury

	Eng_Log("    OK");
	
	Eng_Log("Init life (from file \'Life.ini\') ...");

	Init_Life();
	if((svet = Create_LMatrix("Life.ini")) == NULL)
		return 0;
	// nahraje life

	Eng_Log("    OK");

	Eng_Log("Init engine ...");
	
	Init_Engine(Width, Height);
	Init_Font_Engine(Width, Height);
	// engine ok

	Eng_Log("    OK");

	InitDirectX(hwnd, Width, Height);
	// inicializovano okno

	Eng_Log("DDraw is running..");

	Eng_Log("Locking some memory ...");

	if((buffer = (unsigned __int32*) malloc(Width * Height * 4)) == NULL)
		return 0;

	Eng_Log("    OK");

	Eng_Log("Rendering !");

	SetTimer(hwnd, 0, 20, NULL);

	SetCursorPos(Width >> 1, Height >> 1);
	mx = Width >> 1;
	my = Height >> 1;
	// pripravi MouseLook

	while (GetMessage (&msg, NULL, 0, 0)) {
		TranslateMessage (&msg);
		DispatchMessage (&msg);
	}

	free((void*)buffer);
	// uvolnuje pamet

	Terminate_Engine();
	Terminate_Font_Engine();
	// uvolnuje pamet v jinych modulech

	return msg.wParam;
}

// ati technologies filtr quincunx
void Quincunx(unsigned __int32 *src, unsigned __int32 *dest)
{
	int w2, h2, w1;
	int x, y;

	w2 = Width - 1;
	h2 = Height - 1;
	w1 = w2 + 1;

	y = Width * Height;
	for(x = 0; x < y; x ++)
		src[x] = ((src[x] >> 2) & 0x3f3f3f);

	for(y = 0; y < h2; y ++, dest += Width, src += Width) {
		for(x = 0; x < w2; x ++)
			dest[x] = src[x] + src[x + 1] + src[x + w1] + src[x + w2];
		dest[x] = src[x];
	}
	memcpy(dest, src, Width << 2);

	// obrazovy filtr
}

// nakresli jednu bunku na pozici [x, y, z]
void Draw_Cell(int x, int y, int z)
{
	Matrix m;
	int i;

	x -= x_Size / 2;
	y -= y_Size / 2;
	z -= z_Size / 2;

	Matrix_Init(&m);
	m[3][0] = (float)x * 100;
	m[3][1] = (float)y * 100;
	m[3][2] = (float)z * 100;
	// vytvori matici, ktera urcuje pozici bunky,
	// cela matrice bunek ma stred v [0,0,0], proto "-= 32"

	for(i = 0; i < cell->n_object_num; i ++)
		DrawObject(&m, &cmatrix, &cell->object[i], cell->material);
	// vykresli objekty bunky
}

// nakresli cely prostor simulace
void Draw_Life()
{
	int x, y, z;
	int i;

	for(i = 0; i < Max_Index; i ++) {
		x = (i & x_Mask) >> x_Shift;
		y = (i & y_Mask) >> y_Shift;
		z = (i & z_Mask) >> z_Shift;
		// spocita x, y, z - je to rychlejsi
		// nez tri cykly v sobe ;-)

		if((*svet)[x][y][z] == CELL_LIVE)
			Draw_Cell(x, y, z);
		// kresli zive bunky
	}
}

// funkce, zpracovavajici zpravy
// pri sipkach otaci kamerou, ins/del
// hybe kamerou dopredu dozadu
// a kresli obraz
// pri esc. skonci
LRESULT CALLBACK Wokno (HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
	static float a = 0;
	char str[256];
	POINT p;
	int i;

	switch (iMsg) {
		case WM_KEYDOWN :
			if(wParam == 27){
				ShowCursor(TRUE);
				SendMessage(hwnd, WM_CLOSE, 0, 0);
				return 0;
			} else {
				switch (wParam){
				case VK_RIGHT:
					Matrix_Translate(&_no_y_cmatrix, 30, 0, 0);
					break;
				case VK_LEFT:
					Matrix_Translate(&_no_y_cmatrix, -30, 0, 0);
					break;
				case VK_UP:
					Matrix_Translate(&_no_y_cmatrix, 0, -30 *
						(float)sin(y_angle), 30 * (float)cos(y_angle));
					break;
				case VK_DOWN:
					Matrix_Translate(&_no_y_cmatrix, 0, 30 *
						(float)sin(y_angle), -30 * (float)cos(y_angle));
					// ty kosinusy a sinusy zpusobi ze kamera se muze hybat i nahoru / dolu
					// tudiz uz je to presne jako kdyz v Half-Life nebo kdekoliv date "noclip"
					// (pokud by se rotovala jedinna matice kamery okolo osy x, tak by se mohla
					// otacet v ose pohledu, takze za chvili je vzhuru nohama - pokud smejdite
					// mysi porad dokolecka - a to nechci(byt vzhuru nohama))
					break;
				case VK_HOME:
					have_to_graph = 1;
					break;
				case VK_F12:
					have_to_screenshot = 1;
					break;
				case VK_INSERT:
					Proceed_Life(svet);
					nFrame ++;
					break;
				case VK_DELETE:
					auto_proceed = !auto_proceed;
					break;
				case VK_END:
					bilerp = !bilerp;
					Switch_Bilerp(bilerp);
					break;
				default:
					break;
				}
			} // posouvani kamery sipkama
			return 0;

		case WM_TIMER :
			GetCursorPos(&p);
			SetCursorPos(Width >> 1, Height >> 1);
			ax = p.x - mx;
			ay = p.y - my;
			Matrix_Rotate_Y(&_no_y_cmatrix, (float)ax * (float)3/900);
			y_angle -= (float)ay * (float)3/900;
			memcpy(&cmatrix, &_no_y_cmatrix, sizeof(Matrix));
			Matrix_Rotate_X(&cmatrix, y_angle);
			// osetri mouselook

			memset(buffer, 0, 4 * Width * Height);
			// vycisti obrazovku

			if(auto_proceed) {
				Proceed_Life(svet);
				nFrame ++;
			}
			// hraje life

			Init_Screen();

			Draw_Life();
			// kresli svet

			Release_Screen(buffer);

			if(have_to_graph) {
				Generate_Scanline_Graph();
				have_to_graph = 0;
			}

			if(nFrame % 10 == 1)
					sprintf(str, "%dst Iteration", nFrame);
			else if(nFrame % 10 == 2)
				sprintf(str, "%dnd Iteration", nFrame);
			else
				sprintf(str, "%dth Iteration", nFrame);
			String_Out_AddTr(str, 10, 10, buffer);
			// vypise cislo snimku

			String_Out_AddTr("Press Ins to iterate\n"
					         "Press Del to auto-iterate", 10, Height - 30, buffer);

			if(InitVideoBuffer(&video)) {
				///
				memcpy(video, buffer, 4 * Width * Height);
				//Quincunx(buffer, video);
				/// zkopruje obraz
				FlipVideoBuffer();
			}

			if(have_to_screenshot) {
				for(i = 0; i < Width * Height; i ++) {
					buffer[i] = (buffer[i] & 0x00ff00) |
						       ((buffer[i] >> 16) & 0xff) |
							   ((buffer[i] & 0xff) << 16);
				}
				// prohodi R a B, G zustava na stejnem miste
				Make_ScreenShot(buffer);
				have_to_screenshot = 0;
			}
			// udela screenshot (je potreba prevest RGB->BGR)
			return 0;

		case WM_DESTROY :
			TerminateDirectX();
			PostQuitMessage (0);
			return 0;
	}
	return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
}





