/*
								+---------------------------------+
								|                                 |
								|     ***   Targa reader   ***    |
								|                                 |
								|    Copyright  L. Polok  2001   |
								|                                 |
								|             Tga.cpp             |
								|                                 |
								+---------------------------------+
*/

#include <stdlib.h>
#include <stdio.h>
#include "Tga.h"

typedef struct {
	unsigned char id_length;
	unsigned char palette_type;
	unsigned char image_type;
	unsigned short first_color;
	unsigned short palette_colors;
	unsigned char palette_entry_size;
	unsigned short left;
	unsigned short top;
	unsigned short image_width;
	unsigned short image_height;
	unsigned char bpp;
	unsigned char descriptor_bits;
} TGA_HEADER;

// reads 256color tga with rle
void rle_tga_read(unsigned __int32 *b, int w, FILE *f)
{
	unsigned char value;
	unsigned char count;
	int c = 0;
	int i;

	do {
		fread(&count, sizeof(char), 1, f);
		//
		if(count & 0x80) {
			count = (count & 0x7F) + 1;
			c += count;
			fread(&value, sizeof(value), 1, f);
			while(count --) {
				b[0] = (unsigned __int32)value;
				b ++;
			}
		} else {
			count ++;
			c += count;
			for(i = 0; i < count; i ++) {
				fread(&value, sizeof(value), 1, f);
				b[0] = (unsigned __int32)value;
				b ++;
			}
		}
	} while (c < w);
}

void rle_tga_read2(unsigned char *b, int w, FILE *f)
{
	unsigned char value;
	unsigned char count;
	int c = 0;
	int i;

	do {
		fread(&count, sizeof(char), 1, f);
		//
		if(count & 0x80) {
			count = (count & 0x7F) + 1;
			c += count;
			fread(&value, sizeof(value), 1, f);
			while(count --) {
				b[0] = value;
				b ++;
			}
		} else {
			count ++;
			c += count;
			for(i = 0; i < count; i ++) {
				fread(&value, sizeof(value), 1, f);
				b[0] = value;
				b ++;
			}
		}
	} while (c < w);
}

// reads 32bpp (rgba) tga with rle
void rle_tga_read32(unsigned __int32 *b, int w, FILE *f)
{
	unsigned __int32 value;
	unsigned char count;
	int c = 0;

	do {
		fread(&count, sizeof(char), 1, f);
		//
		if(count & 0x80) { // Most Significant Bit = rle mark
			count = (count & 0x7F) + 1;
			c += count;
			fread(&value, sizeof(value), 1, f);
			while(count --)
				*b ++ = value;
		} else {
			count ++;
			c += count;
			while(count --) {
				fread(&value, sizeof(value), 1, f);
				*b ++ = value;
			}
		}
	} while (c < w);
}

// reads 24bpp tga with rle
void rle_tga_read24(unsigned __int32 *b, int w, FILE *f)
{
	unsigned char value[3];
	unsigned char count;
	int c = 0;

	do {
		fread(&count, sizeof(char), 1, f);
		//
		if(count & 0x80) { // MSB = rle mark
			count = (count & 0x7F) + 1;
			c += count;
			fread(value, sizeof(value), 1, f);
			while(count --)
				*b ++ = (value[0] << 16) + (value[1] << 8) + value[2];
		} else {
			count ++;
			c += count;
			while(count --) {
				fread(value, sizeof(value), 1, f);
				*b ++ = (value[0] << 16) + (value[1] << 8) + value[2];
			}
		}
	} while (c < w);
}

// reads 16 color tga with rle (but g = 5 bites ???)
void rle_tga_read16(unsigned __int32 *b, int w, FILE *f)
{
	unsigned short value;
	unsigned __int32 color;
	unsigned char count;
	int c = 0;

	do {
		fread(&count, sizeof(char), 1, f);
		//
		if(count & 0x80) { // MSB = rle mark
			count = (count & 0x7f) + 1;
			c += count;
			fread(&value, sizeof(short), 1, f);
			color = (((value >> 10) & 0x1f) << 3) +
					(((value >> 5) & 0x1f) << 11) +
					((value & 0x1f) << 19); // repacks 16bpp -> 32bpp
			while(count --)
			    *(b ++) = color;
		} else {
			count ++;
			c += count;
			while(count --) {
				fread(&value, sizeof(short), 1, f);
				color = (((value >> 10) & 0x1f) << 3) +
						(((value >> 5) & 0x1f) << 11) +
						((value & 0x1f) << 19);
				*(b ++) = color;
			}
		}
	} while(c < w);
}

// reads targa header
TGA_HEADER Read_TGA_Header(FILE *f)
{
	TGA_HEADER TGAHead;

	fread(&TGAHead.id_length, 1, 1, f);
	fread(&TGAHead.palette_type, 1, 1, f);
	fread(&TGAHead.image_type, 1, 1, f);
	fread(&TGAHead.first_color, 2, 1, f);
	fread(&TGAHead.palette_colors, 2, 1, f);
	fread(&TGAHead.palette_entry_size, 1, 1, f);
	fread(&TGAHead.left, 2, 1, f);
	fread(&TGAHead.top, 2, 1, f);
	fread(&TGAHead.image_width, 2, 1, f);
	fread(&TGAHead.image_height, 2, 1, f);
	fread(&TGAHead.bpp, 1, 1, f);
	fread(&TGAHead.descriptor_bits, 1, 1, f);

	return TGAHead;
}

TGA *Load_Pal_TGA(char *filename)
{
	unsigned char image_id[256], image_palette[256][3];
	unsigned char rgb[3], rgbq[4];
	unsigned short rgbs[1];
	TGA_HEADER TGAHead;
	unsigned char c[1];
	int i, x, y, yc;
	int compressed;
	FILE *TGAfp;
	TGA *tga;

	if((TGAfp = fopen(filename, "rb")) == NULL)
		return NULL;
	// opens file

	TGAHead = Read_TGA_Header(TGAfp);
	// fread(&TGAHead, 1, sizeof(TGAHead), TGAfp);
	//    ^- impossible due to structure memory align
	// reads header

	fread(image_id, 1, TGAHead.id_length, TGAfp);
	// reads image id

	for(i = 0; i < TGAHead.palette_colors; i ++) {
		switch(TGAHead.palette_entry_size) {
		case 16: 
			fread(rgbs, sizeof(rgbs), 1, TGAfp);
			image_palette[i][0] = (rgbs[0] & 0x1F) << 3;
			image_palette[i][1] = ((rgbs[0] >> 5) & 0x1F) << 3;
			image_palette[i][2] = ((rgbs[0] >> 10) & 0x1F) << 3;
			//
			break;
		case 24:
			fread(rgb, sizeof(rgb), 1, TGAfp);
			image_palette[i][0] = rgb[0];
			image_palette[i][1] = rgb[1];
			image_palette[i][2] = rgb[2];
			//
			break;
		case 32:
			fread(rgbq, sizeof(rgbq), 1, TGAfp);
			image_palette[i][0] = rgbq[0];
			image_palette[i][1] = rgbq[1];
			image_palette[i][2] = rgbq[2];
			//
			break;
		}
	}
	// reads palette

	if(TGAHead.image_type == 3) {
		for (i=0; i < 256; i ++) {
			image_palette[i][0] = i;
			image_palette[i][1] = i;
			image_palette[i][2] = i;
		}
	}
	// grayscale pal

	/*/ Image type:
	 *    0 = no image data
	 *    1 = uncompressed color mapped
	 *    2 = uncompressed true color
	 *    3 = grayscale
	 *    9 = RLE color mapped
	 *   10 = RLE true color
	 *   11 = RLE grayscale
	/*/

	compressed = (TGAHead.image_type & 0x08);
	TGAHead.image_type &= 0x07;
	// do some calculations

	if((TGAHead.image_type < 1) || (TGAHead.image_type > 3)) {
		fclose(TGAfp);
		return NULL;
	}
	// check TGA

	if(TGAHead.image_type == 2) {
		fclose(TGAfp);
		return NULL;
	}
	// non-palette TGA

	if((tga = (TGA*) malloc(sizeof(TGA))) == NULL) {
		fclose(TGAfp);
		return NULL;
	}
	if((tga->bytes = (unsigned char*) malloc
	   (TGAHead.image_width * TGAHead.image_height)) == NULL) {
		fclose(TGAfp);
		return NULL;
	}
	if((tga->palette = (unsigned __int32*) malloc
	   (TGAHead.palette_colors * 4)) == NULL) {
		fclose(TGAfp);
		return NULL;
	}
	tga->Width = TGAHead.image_width;
	tga->Height = TGAHead.image_height;
	// create bitmap

	for(x = 0; x < TGAHead.palette_colors; x ++) {
		tga->palette[x] = (image_palette[x][2] << 16) +
						  (image_palette[x][1] << 8) +
					       image_palette[x][0];
	}
	tga->color_num = TGAHead.palette_colors;
	// create palette

	for(y = TGAHead.image_height; y > 0; y --) {
		yc = (TGAHead.descriptor_bits & 0x20) ? TGAHead.image_height - y : y - 1;
		// compute real y

		switch(TGAHead.image_type) {
		case 1:
		case 3:
			if(compressed) {
				rle_tga_read2(&tga->bytes[yc * tga->Width], tga->Width, TGAfp);
			} else {
				for(x = 0; x < TGAHead.image_width; x ++) {
					fread(c, sizeof(c), 1, TGAfp);
					tga->bytes[yc * tga->Width + x] = c[0];
				}
			}
			break;
		}
	}
	// reads image

	fclose(TGAfp);
	// close file

	return tga;
}

// nahraje targa obrazek
// zda se ok (16 / 14 / 32 bpp - ale jen nekomprim. zbytek jsem nezkousel)
// (jo.. (ukladal jsem photoshopem - nevim jak ho prinutit k rle kompresi))
// inac to chi davat jeste jinam - proto english komentare :-)
/*BMP *Load_TGA(char *filename)
{
	unsigned char image_id[256], image_palette[256][3];
	unsigned char rgb[3], rgbq[4];
	unsigned short rgbs[1];
	TGA_HEADER TGAHead;
	unsigned char c[1];
	int i, x, y, yc;
	int compressed;
	FILE *TGAfp;
	BMP *bmp;

	int mx_color = 0;

	if((TGAfp = fopen(filename, "rb")) == NULL)
		return NULL;
	// opens file

	TGAHead = Read_TGA_Header(TGAfp);
	// fread(&TGAHead, 1, sizeof(TGAHead), TGAfp);
	//    ^- impossible due to structure memory align
	// reads header

	fread(image_id, 1, TGAHead.id_length, TGAfp);
	// reads image id

	for(i = 0; i < TGAHead.palette_colors; i ++) {
		switch(TGAHead.palette_entry_size) {
		case 16: 
			fread(rgbs, sizeof(rgbs), 1, TGAfp);
			image_palette[i][0] = (rgbs[0] & 0x1F) << 3;
			image_palette[i][1] = ((rgbs[0] >> 5) & 0x1F) << 3;
			image_palette[i][2] = ((rgbs[0] >> 10) & 0x1F) << 3;
			//
			break;
		case 24:
			fread(rgb, sizeof(rgb), 1, TGAfp);
			image_palette[i][0] = rgb[0];
			image_palette[i][1] = rgb[1];
			image_palette[i][2] = rgb[2];
			//
			break;
		case 32:
			fread(rgbq, sizeof(rgbq), 1, TGAfp);
			image_palette[i][0] = rgbq[0];
			image_palette[i][1] = rgbq[1];
			image_palette[i][2] = rgbq[2];
			//
			break;
		}
	}
	// reads palette

	if(TGAHead.image_type == 3) {
		for (i=0; i < 256; i ++) {
			image_palette[i][0] = i;
			image_palette[i][1] = i;
			image_palette[i][2] = i;
		}
	}
	// grayscale pal

	/* / Image type:
	 *    0 = no image data
	 *    1 = uncompressed color mapped
	 *    2 = uncompressed true color
	 *    3 = grayscale
	 *    9 = RLE color mapped
	 *   10 = RLE true color
	 *   11 = RLE grayscale
	/* /

	compressed = (TGAHead.image_type & 0x08);
	TGAHead.image_type &= 0x07;
	// do some calculations

	if((TGAHead.image_type < 1) || (TGAHead.image_type > 3)) {
		fclose(TGAfp);
		return NULL;
	}
	// check

	if((bmp = (BMP*) malloc(sizeof(BMP))) == NULL) {
		fclose(TGAfp);
		return NULL;
	}
	if((bmp->bytes = (unsigned __int32*) malloc
	   (TGAHead.image_width * TGAHead.image_height * 4)) == NULL) {
		fclose(TGAfp);
		return NULL;
	}
	bmp->width = TGAHead.image_width;
	bmp->height = TGAHead.image_height;
	// create bitmap

	for(y = TGAHead.image_height; y > 0; y --) {
		yc = (TGAHead.descriptor_bits & 0x20) ? TGAHead.image_height - y : y - 1;
		// compute real y

		switch(TGAHead.image_type) {
		case 1:
		case 3:
			if(compressed) {
				rle_tga_read(&bmp->bytes[yc * bmp->width], bmp->width, TGAfp);
				for(x = 0; x < TGAHead.image_width; x ++) {
					bmp->bytes[yc * bmp->width + x] =
						(image_palette[bmp->bytes[yc * bmp->width + x]][0] << 16) +
						(image_palette[bmp->bytes[yc * bmp->width + x]][1] << 8) +
						image_palette[bmp->bytes[yc * bmp->width + x]][2];
				}
			} else {
				for(x = 0; x < TGAHead.image_width; x ++) {
					fread(c, sizeof(c), 1, TGAfp);
					if(c[0] > mx_color)
						mx_color = c[0];
					bmp->bytes[yc * bmp->width + x] =
						(image_palette[c[0]][0] << 16) +
						(image_palette[c[0]][1] << 8) +
						image_palette[c[0]][2];
				}
			}
			break;

		case 2:
			if(TGAHead.bpp == 32) {
				if(compressed) {
					rle_tga_read32(&bmp->bytes[yc * bmp->width], bmp->width, TGAfp);
				} else {
					for(x = 0; x < TGAHead.image_width; x ++) {
						fread(rgbq, sizeof(rgbq), 1, TGAfp);
						bmp->bytes[yc * bmp->width + x] = (rgbq[3] << 24) +
							(rgbq[0] << 16) + (rgbq[1] << 8) + rgbq[2];
					}
				}
			} else if(TGAHead.bpp == 24) {
				if(compressed) {
					rle_tga_read24(&bmp->bytes[yc * bmp->width], bmp->width, TGAfp);
				} else {
					for(x = 0; x < TGAHead.image_width; x ++) {
						fread(rgb, sizeof(rgb), 1, TGAfp);
						bmp->bytes[yc * bmp->width + x] = (rgb[0] << 16)
							+ (rgb[1] << 8) + rgb[2];
					}
				}
			} else /* if(TGAHead.bpp == 16) { * / {
				if(compressed) {
					rle_tga_read16(&bmp->bytes[yc * bmp->width], bmp->width, TGAfp);
				} else {
					for(x = 0; x < TGAHead.image_width; x ++) {
						fread(rgbs, sizeof(rgbs), 1, TGAfp);
						/*r = ((rgbs & 0x7c00) >> 7) & 0xff;
						g = ((rgbs & 0x03e0) >> 2) & 0xff; 
						b = ((rgbs & 0x001f) << 3) & 0xff;* /
						bmp->bytes[yc * bmp->width + x] =
							(((rgbs[0] >> 10) & 0x1f) << 3) +
							(((rgbs[0] >> 5) & 0x1f) << 11) +
							((rgbs[0] & 0x1f) << 19);
					}
				}
			}
			break;
		}
	}
	// reads image

	fclose(TGAfp);
	// close file

	return bmp;
}*/